import json
from pedurma.exceptions import TextMappingNotFound
import re
import requests

from collections import defaultdict
from pathlib import Path
from typing import List, Optional

from openpecha.cli import download_pecha
from openpecha.serializers import HFMLSerializer

from pedurma.pecha import *
from pedurma.utils import from_yaml




def get_text_info(text_id, index):
    texts = index["annotations"]
    for uuid, text in texts.items():
        if text["work_id"] == text_id:
            return (uuid, text)
    return ("", "")


def get_meta_data(pecha_id, text_uuid, meta_data):
    meta = {}
    source_meta = meta_data.get("source_metadata", "")
    if source_meta:
        meta = {
            "work_id": source_meta.get("work_id", ""),
            "img_grp_offset": source_meta.get("img_grp_offset",""),
            "pref": source_meta.get("pref", ""),
            "pecha_id": pecha_id,
            "text_uuid": text_uuid,
        }
    return meta


def get_hfml_text(opf_path, text_id, index=None):
    serializer = HFMLSerializer(opf_path, text_id=text_id, index_layer=index, layers=['Pagination', 'Durchen'])
    serializer.apply_layers()
    hfml_text = serializer.get_result()
    return hfml_text

def get_prev_pg_ann(pg_num, pg_face):
    prev_pg_ann = '['
    if pg_face == 'a':
        prev_pg_ann += f'{pg_num-1}b]'
    else:
        prev_pg_ann += f'{pg_num}b]'
    return prev_pg_ann

def add_first_page_ann(text):
    pg_pat = re.search(r"\[[𰵀-󴉱]?([0-9]+)([a-z]{1})\]", text)
    pg_num = int(pg_pat.group(1))
    pg_face = pg_pat.group(2)
    prev_pg_ann = get_prev_pg_ann(pg_num, pg_face)
    new_text = f'{prev_pg_ann}\n{text}'
    return new_text


def get_body_text(text_with_durchen):
    try:
        durchen_starting = re.search("<[𰵀-󴉱]?d", text_with_durchen).start()
        text_content = text_with_durchen[:durchen_starting]
    except Exception:
        text_content = text_with_durchen
    return text_content


def get_durchen(text_with_durchen):
    durchen = ""
    durchen_start = False
    pages = get_pages(text_with_durchen)
    for page in pages:
        if re.search("<[𰵀-󴉱]?d", page) or durchen_start == True:
            durchen += page
            durchen_start = True
        if re.search('d>', page):
            return durchen
    if not durchen:
        print('INFO: durchen not found..')
    return durchen


def get_pages(vol_text):
    result = []
    pg_text = ""
    pages = re.split(r"(\[[𰵀-󴉱]?[0-9]+[a-z]{1}\])", vol_text)
    for i, page in enumerate(pages[1:]):
        if i % 2 == 0:
            pg_text += page
        else:
            pg_text += page
            result.append(pg_text)
            pg_text = ""
    return result


def get_page_id(page_idx, pagination_layer):
    paginations = pagination_layer["annotations"]
    for uuid, pagination in paginations.items():
        if pagination["page_index"] == page_idx:
            return (uuid, pagination)
    return ("", "")


def get_page_num(page_ann):
    pg_num = int(page_ann[:-1]) * 2
    pg_face = page_ann[-1]
    if pg_face == "a":
        pg_num -= 1
    return pg_num


def get_link(pg_num, text_meta):
    vol = text_meta["vol"]
    img_group_offset = text_meta["img_grp_offset"]
    pref = text_meta["pref"]
    igroup = f"{pref}{img_group_offset+vol}"
    link = f"https://iiif.bdrc.io/bdr:{igroup}::{igroup}{int(pg_num):04}.jpg/full/max/0/default.jpg"
    return link


def get_note_ref(pagination):
    try:
        return pagination["note_ref"]
    except Exception:
        return ""


def get_clean_page(page):
    pat_list = {
            "page_pattern": r"\[([𰵀-󴉱])?[0-9]+[a-z]{1}\]",
            "topic_pattern": r"\{([𰵀-󴉱])?\w+\}",
            "start_durchen_pattern": r"\<([𰵀-󴉱])?d",
            "end_durchen_pattern": r"d\>",
            "sub_topic_pattern": r"\{([𰵀-󴉱])?\w+\-\w+\}",
        }
    base_page = page
    for ann, ann_pat in pat_list.items():
        base_page = re.sub(ann_pat, "", base_page)
    base_page = base_page.lstrip()
    return base_page


def get_page_obj(page, text_meta, tag, pagination_layer):
    page_idx = re.search(r"\[([𰵀-󴉱])?([0-9]+[a-z]{1})\]", page).group(2)
    page_id, pagination = get_page_id(page_idx, pagination_layer)
    page_content = get_clean_page(page)
    pg_num = get_page_num(page_idx)
    page_link = get_link(pg_num, text_meta)
    note_ref = get_note_ref(pagination)
    if page_content == "":
        page_obj = None
    else:
        if tag == "note":
            page_obj = NotesPage(
                id=page_id,
                page_no=pg_num,
                content=page_content,
                name=f"Page {pg_num}",
                vol=text_meta["vol"],
                image_link=page_link,
            )
        else:
            page_obj = Page(
                id=page_id,
                page_no=pg_num,
                content=page_content,
                name=f"Page {pg_num}",
                vol=text_meta["vol"],
                image_link=page_link,
                note_ref=note_ref,
            )

    return page_obj


def get_page_obj_list(text, text_meta, pagination_layer, tag="text"):
    page_obj_list = []
    pages = get_pages(text)
    for page in pages:
        pg_obj = get_page_obj(page, text_meta, tag, pagination_layer)
        if pg_obj:
            page_obj_list.append(pg_obj)
    return page_obj_list


def construct_text_obj(hfmls, text_meta, opf_path):
    pages = []
    notes = []
    vol_span = []
    for vol_num, hfml_text in hfmls.items():
        text_meta["vol"] = int(vol_num[1:])
        pagination_layer = from_yaml(
            Path(
                f"{opf_path}/{text_meta['pecha_id']}.opf/layers/v{int(text_meta['vol']):03}/Pagination.yml"
            )
        )
        durchen = get_durchen(hfml_text)
        body_text = hfml_text.replace(durchen, '')
        
        pages += get_page_obj_list(body_text, text_meta, pagination_layer, tag="text")
        if durchen:
            notes += get_page_obj_list(durchen, text_meta, pagination_layer, tag="note")
    text_obj = Text(id=text_meta["text_uuid"], pages=pages, notes=notes)
    return text_obj


def serialize_text_obj(text):
    """Serialize text object to hfml

    Args:
        text (obj): text object

    Returns:
        dict: vol as key and value as hfml
    """
    text_hfml = defaultdict(str)
    pages = text.pages
    notes = text.notes
    for page in pages:
        if page.page_no != 1:
            text_hfml[f"v{int(page.vol):03}"] += f'\n{page.content}'
        else:
            text_hfml[f"v{int(page.vol):03}"] += f'{page.content}'
    for note in notes:
        text_hfml[f"v{int(note.vol):03}"] += f'\n{note.content}'
    return text_hfml

def get_durchen_page_obj(page, notes):
    for note in notes:
        if note.id == page.note_ref:
            return note
    return None

def get_pecha_paths(text_id, text_mapping=None):
    pecha_paths = {
        "namsel": None,
        "google": None
    }
    if not text_mapping:
        text_mapping = requests.get('https://raw.githubusercontent.com/OpenPecha-dev/editable-text/main/t_text_list.json')
        text_mapping = json.loads(text_mapping.text)
    text_info = text_mapping.get(text_id, {})
    if text_info:
        pecha_paths['namsel'] = download_pecha(text_info['namsel'])
        pecha_paths['google'] = download_pecha(text_info['google'])
    else:
        raise TextMappingNotFound
    return pecha_paths

def get_text_obj(pecha_id, text_id, pecha_path = None):
    if not pecha_path:
        pecha_path = download_pecha(pecha_id, needs_update=False)
    meta_data = from_yaml(Path(f"{pecha_path}/{pecha_id}.opf/meta.yml"))
    index = from_yaml(Path(f"{pecha_path}/{pecha_id}.opf/index.yml"))
    hfmls = get_hfml_text(f"{pecha_path}/{pecha_id}.opf/", text_id, index)
    text_uuid, text = get_text_info(text_id, index)
    text_meta = get_meta_data(pecha_id, text_uuid, meta_data)
    text = construct_text_obj(hfmls, text_meta, pecha_path)
    return text

def get_pedurma_text_obj(text_id, pecha_paths=None):
    if not pecha_paths:
        pecha_paths = get_pecha_paths(text_id)
    text = {}
    for pecha_src, pecha_path in pecha_paths.items():
        pecha_id = Path(pecha_path).stem
        text[pecha_src] = get_text_obj(pecha_id, text_id, pecha_path)
    pedurma_text = PedurmaText(text_id= text_id, namsel=text['namsel'], google=text['google'])
    return pedurma_text


# if __name__ == "__main__":
#     text_id = 'D1118'
#     pecha_id = 'P000792'
#     opf_path = f'./test/{pecha_id}.opf'
#     index = from_yaml(Path(f"./test/{pecha_id}.opf/index.yml"))
#     meta_data = from_yaml(Path(f"./test/{pecha_id}.opf/meta.yml"))
#     text_uuid, text_info = get_text_info(text_id, index)
#     text_meta = get_meta_data(pecha_id, text_uuid, meta_data)
#     hfmls = get_hfml_text(text_id, opf_path)
#     text_obj = get_text_obj(hfmls, text_meta, opf_path)



import re


def construct_notes(note_id, note):
    combined_notes = f'<({note_id})'
    for pub, note_value in note.items():
        if note_value:
            combined_notes += f"{pub}{note_value},"
    combined_notes += '>'
    return combined_notes


def reinsert_pedurma_notes(pg, notes):
    new_pg = re.sub('[𰵀-󴉱]', '', pg)
    for note_id, note in notes.items():
        combined_note = construct_notes(note_id, note)
        new_pg = re.sub('#', combined_note, new_pg, 1)
    return new_pg




import copy
from openpecha.blupdate import *
from openpecha.cli import download_pecha
from openpecha.utils import load_yaml, dump_yaml
from pedurma.pecha import *
from pedurma.texts import serialize_text_obj, get_pecha_paths
from pedurma.utils import from_yaml, to_yaml

def get_old_vol(pecha_opf_path, pecha_id, text_vol_span):
    """Generate old base text in which text is located

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id 
        text_vol_span (list): list of volume ids in which text is located

    Returns:
        dict: volid as key and base as value
    """
    old_vols = {}
    for vol_id in text_vol_span:
        old_vols[vol_id] = Path(f"{pecha_opf_path}/{pecha_id}.opf/base/{vol_id}.txt").read_text(encoding='utf-8')
    return old_vols

def get_old_text_base(old_pecha_idx, old_vol_base, text_id, text_vol_num):
    """Return old text base

    Args:
        old_pecha_idx (dict): old index
        old_vol_base (dict): old vol base and its vol id
        text_id (str): text id
        text_vol_num (str): text vol id

    Returns:
        str: text basetext in that volume
    """
    text_span = old_pecha_idx['annotations'][text_id]['span']
    for vol_span in text_span:
        if vol_span['vol'] == text_vol_num:
            return old_vol_base[vol_span['start']:vol_span['end']]
    return ''

def get_new_vol(old_vols, old_pecha_idx, text_obj):
    """Return new base text by replacing updated text from text object

    Args:
        old_vols (dict): old vol and its id
        old_pecha_idx (dict): old pecha index
        text_obj (obj): text object

    Returns:
        dict: new basetext and its id
    """
    new_vols = {}
    new_text = serialize_text_obj(text_obj)
    for vol_id, new_text_base in new_text.items():
        vol_num = int(vol_id[1:])
        old_vol_base = old_vols[vol_id]
        old_text_base = get_old_text_base(old_pecha_idx, old_vol_base, text_obj.id, vol_num)
        old_text_base = old_text_base.strip()
        new_text_base = new_text_base.strip()
        new_vol_base = old_vol_base.replace(old_text_base, new_text_base)
        new_vols[vol_id] = new_vol_base
    return new_vols

def get_text_vol_span(pecha_idx, text_uuid):
    """Return list of volume ids in which text span

    Args:
        pecha_idx (dict): pecha index
        text_uuid (uuid): text uuid

    Returns:
        list: vol ids
    """
    text_vol_span = []
    for span in pecha_idx['annotations'][text_uuid]['span']:
        vol_num = span['vol']
        text_vol_span.append(f'v{int(vol_num):03}')
    return text_vol_span

def update_base(pecha_opf_path, pecha_id, text_obj, old_pecha_idx):
    """Update base text using text obj

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        text_obj (obj): text object
        old_pecha_idx (dict): old pecha index
    """
    text_vol_span = get_text_vol_span(old_pecha_idx, text_obj.id)
    old_vols = get_old_vol(pecha_opf_path, pecha_id, text_vol_span)
    new_vols = get_new_vol(old_vols, old_pecha_idx, text_obj)
    for vol_id, new_vol_base in new_vols.items():
        Path(f"{pecha_opf_path}/{pecha_id}.opf/base/{vol_id}.txt").write_text(new_vol_base, encoding='utf-8')
        print(f'INFO: {vol_id} base updated..')

def get_old_layers(pecha_opf_path, pecha_id, vol_id):
    """Return all the layers belonging in volume

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        vol_id (str): volume id

    Returns:
        dict: layer name as key and layer annotations as value
    """
    old_layers = {}
    layer_paths = list(Path(f"{pecha_opf_path}/{pecha_id}.opf/layers/{vol_id}").iterdir())
    for layer_path in layer_paths:
        layer_name = layer_path.stem
        layer_content = from_yaml(layer_path)
        old_layers[layer_name] = layer_content
    return old_layers

def update_layer(pecha_opf_path, pecha_id, vol_id, old_layers, updater):
    """Update particular layers belonging in given volume id 

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        vol_id (str): volume id
        old_layers (dict): layer name as key and annotations as value
        updater (obj): updater object
    """
    for layer_name, old_layer in old_layers.items():
        update_ann_layer(old_layer, updater)
        new_layer_path = Path(f"{pecha_opf_path}/{pecha_id}.opf/layers/{vol_id}/{layer_name}.yml")
        dump_yaml(old_layer, new_layer_path)
        print(f'INFO: {vol_id} {layer_name} has been updated...')
    
def update_old_layers(pecha_opf_path, pecha_id, text_obj, old_pecha_idx):
    """Update all the layers related to text object

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        text_obj (obj): text object
        old_pecha_idx (dict): old pecha index
    """
    text_vol_span = get_text_vol_span(old_pecha_idx, text_obj.id)
    old_vols = get_old_vol(pecha_opf_path, pecha_id, text_vol_span)
    new_vols = get_new_vol(old_vols, old_pecha_idx, text_obj)
    for (vol_id, old_vol_base), (_, new_vol_base) in zip(old_vols.items(), new_vols.items()):
        updater = Blupdate(old_vol_base, new_vol_base)
        old_layers = get_old_layers(pecha_opf_path, pecha_id, vol_id)
        update_layer(pecha_opf_path, pecha_id, vol_id, old_layers, updater)

def update_other_text_index(old_pecha_idx, text_id, cur_vol_offset, vol_num):
    check_flag = False
    for text_uuid, text in old_pecha_idx['annotations'].items():
        if check_flag:
            for vol_walker, vol_span in enumerate(text['span']):
                if vol_span['vol'] == vol_num:
                    old_pecha_idx["annotations"][text_uuid]['span'][vol_walker]['start'] += cur_vol_offset
                    old_pecha_idx["annotations"][text_uuid]['span'][vol_walker]['end'] += cur_vol_offset
                elif vol_span['vol'] > vol_num:
                    return old_pecha_idx
        if text_uuid == text_id:
            check_flag = True
    return old_pecha_idx

def update_index(pecha_opf_path, pecha_id, text_obj, old_pecha_idx):
    """Update pecha index according to text obj content

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        text_obj (obj): text object
        old_pecha_idx (dict): old pecha index

    Returns:
        dict: new pecha index
    """
    text_vol_span = get_text_vol_span(old_pecha_idx, text_obj.id)
    old_vols = get_old_vol(pecha_opf_path, pecha_id, text_vol_span)
    new_vols = get_new_vol(old_vols, old_pecha_idx, text_obj)
    for (vol_id, old_vol_base), (_, new_vol_base) in zip(old_vols.items(), new_vols.items()):
        check_next_text = True
        vol_num = int(vol_id[1:])
        cur_vol_offset = len(new_vol_base) - len(old_vol_base)
        if cur_vol_offset != 0:
            for vol_walker, vol_span in enumerate(old_pecha_idx["annotations"][text_obj.id]['span']):
                if vol_span['vol'] == vol_num:
                    old_pecha_idx["annotations"][text_obj.id]['span'][vol_walker]['end'] += cur_vol_offset
                elif vol_span['vol'] > vol_num:
                    check_next_text = False
                    break
            if check_next_text:
                old_pecha_idx = update_other_text_index(old_pecha_idx, text_obj.id, cur_vol_offset, vol_num)
    return old_pecha_idx

def save_text(pecha_id, text_obj, pecha_opf_path=None, **kwargs):
    """Update pecha opf according to text object content

    Args:
        pecha_id (str): pecha id
        text_obj (text obj): text object
        pecha_opf_path (str, optional): pecha path. Defaults to None.

    Returns:
        path: pecha opf path
    """
    if not pecha_opf_path:
        pecha_opf_path = download_pecha(pecha_id, **kwargs)
    old_pecha_idx = from_yaml(Path(f'{pecha_opf_path}/{pecha_id}.opf/index.yml'))
    prev_pecha_idx = copy.deepcopy(old_pecha_idx)
    new_pecha_idx = update_index(pecha_opf_path, pecha_id, text_obj, old_pecha_idx)
    update_old_layers(pecha_opf_path, pecha_id, text_obj, prev_pecha_idx)
    update_base(pecha_opf_path, pecha_id, text_obj, prev_pecha_idx)
    new_pecha_idx_path = Path(f'{pecha_opf_path}/{pecha_id}.opf/index.yml')
    dump_yaml(new_pecha_idx, new_pecha_idx_path)
    return pecha_opf_path

def get_pedurma_text_mapping(pedurma_text_obj):
    """Pedurma text obj are parse and added pecha path

    Args:
        pedurma_text_obj (obj): pedurma text obj

    Returns:
        dict: ocr engine as key and associated text data as value
    """
    pedurma_text_mapping = {}
    pecha_paths = get_pecha_paths(text_id=pedurma_text_obj.text_id)
    for pecha_src, pecha_path in pecha_paths.items():
        if pecha_src == 'namsel':
            text_obj = pedurma_text_obj.namsel
        else:
            text_obj = pedurma_text_obj.google
        pedurma_text_mapping[pecha_src] = {
            'pecha_id': Path(pecha_path).stem,
            'text_obj': text_obj,
            'pecha_path': pecha_path
        }
    return pedurma_text_mapping

def save_pedurma_text(pedurma_text_obj, pedurma_text_mapping=None):
    """Save changes to respective pedurma opfs according to pedurma text object content

    Args:
        pedurma_text_obj (obj): pedurma text object
        pedurma_text_mapping (dict, optional): pedurma text data mapping. Defaults to None.
    """
    if not pedurma_text_mapping:
        pedurma_text_mapping = get_pedurma_text_mapping(pedurma_text_obj)
    for ocr_engine, pedurma_text in pedurma_text_mapping.items():
        save_text(pedurma_text['pecha_id'], pedurma_text['text_obj'], pedurma_text['pecha_path'])


class PageNumMissing(BaseException):

     def __init__(self):
        self.message = f"page number not found in note pages. Please correct the page num of preview page in note pages.."

class TextMappingNotFound(BaseException):

     def __init__(self):
        self.message = f"Text mapping to its opf is not found. Please text pecha mapping.json in editable text repo"


import yaml

from pathlib import Path
from pydantic.networks import AnyHttpUrl

from pedurma.utils import from_yaml, to_yaml
from pedurma.pecha import PageWithNote
from pedurma.reinsertion import reinsert_pedurma_notes



class Pedurma:

    def __init__(self, project_name:str, base_path: Path):
        self.project_name = project_name
        self.base_path = base_path / project_name
        self.meta = {
        "work_id": 'W1PD95844',
        "img_grp_offset": 845,
        "pref": 'I1PD95'
        } #tenjur meta
    
    def get_vol_num(self, text_id:str)->int:
        text_infos = text_id.split('_')
        vol_num = text_infos[1]
        return int(vol_num)

    def get_page_meta(self, text_id: str, page_id:str)->dict:
        page_meta = {}
        vol_num = self.get_vol_num(text_id)
        page_meta['vol'] = vol_num
        page_meta['page_num'] = page_id
        return page_meta

    def get_page_content(self, text_id: str, page_id:str)->str:
        page_content = (self.base_path / f"{text_id}/pages/{page_id}.txt").read_text(encoding='utf-8')
        return page_content
    
    def get_page_note(self, text_id: str, page_id:str)->str:
        page_content = from_yaml((self.base_path / f"{text_id}/notes/{page_id}.yml"))
        return page_content
    
    def get_image_link(self, page_meta:dict)->AnyHttpUrl:
        vol = page_meta["vol"]
        pg_num = page_meta['page_num']
        img_group_offset = self.meta["img_grp_offset"]
        pref = self.meta["pref"]
        igroup = f"{pref}{img_group_offset+vol}"
        link = f"https://iiif.bdrc.io/bdr:{igroup}::{igroup}{int(pg_num):04}.jpg/full/max/0/default.jpg"
        return link

    def get_page(self, text_id: str, page_id:str) -> PageWithNote:
        page_meta = self.get_page_meta(text_id, page_id)
        page_content = self.get_page_content(text_id, page_id)
        page_image_link = self.get_image_link(page_meta)
        page_note = self.get_page_note(text_id, page_id)
        note = to_yaml(page_note['notes'])
        note_image_link = page_note['note_pg_link']
        page = PageWithNote(
            text_id = text_id,
            page_id = page_id,
            content= page_content,
            page_image_link= page_image_link,
            note=note,
            note_image_link=note_image_link
        )
        return page
    
    def save_page(self,page: PageWithNote):
        text_id = page.text_id
        page_id = page.page_id
        # Page content save
        new_page_content = page.content
        (self.base_path / f"{text_id}/pages/{page_id}.txt").write_text(new_page_content, encoding='utf-8')
        # Note content save
        new_note = yaml.load(page.note, Loader=yaml.CLoader)
        page_note = self.get_page_note(text_id, page_id)
        page_note['notes'] = new_note
        new_page_note = to_yaml(page_note)
        (self.base_path / f"{text_id}/notes/{page_id}.yml").write_text(new_page_note, encoding='utf-8')

    def get_preview(self, page: PageWithNote) -> str:
        pg_content = page.content
        notes = yaml.load(page.note, Loader=yaml.CLoader)
        preview_pg = reinsert_pedurma_notes(pg_content, notes)
        return preview_pg
    
    def get_collation_text(self, text_id: str) -> str:
        collation_text = ''
        page_ids = list((self.base_path / f"{text_id}/notes").iterdir())
        page_ids.sort()
        page_ids = [page_id.stem for page_id in page_ids]
        for page_id in page_ids:
            page_obj = self.get_page(text_id, page_id)
            collation_text += f'{self.get_preview(page_obj)}\n\n'
        return collation_text

